/* SPDX-License-Identifier: LGPL-2.1-or-later */
.file-editor-modal, .file-create-modal {
  block-size: 100%;

  &.is-modified {
    /* Add a background highlight on the modal itself */
    &::after {
      background: linear-gradient(to bottom, var(--pf-t--global--icon--color--status--warning--default) 0.33em, transparent 0.33em);
      content: '';
      inset: 0;
      pointer-events: none;
      position: absolute;
    }

    /* Add a circle to indicate that there are changes */
    .pf-v6-c-modal-box__title-text::after {
      aspect-ratio: 1;
      background: var(--pf-t--global--icon--color--status--warning--default);
      block-size: 1cap;
      border-radius: 0.5cap;
      content: '';
      display: inline-block;
      margin-inline-start: 0.5cap;
    }
  }
  /* Create mode shows textarea as full size */
  .pf-v6-c-form.pf-m-horizontal {
    block-size: 100%;
    grid-template-rows: auto 1fr auto;
  }
}

.file-editor {
  block-size: 100%;

  textarea {
    font-family: var(--pf-t--global--font--family--mono);
    hyphens: none;
    white-space: wrap break-spaces;
  }
}

.file-editor-title-label {
  margin-inline-start: var(--pf-t--global--spacer--sm);
}
