/*
 * Copyright (C) 2016 Red Hat, Inc.
 * SPDX-License-Identifier: LGPL-2.1-or-later
 */

/**
 This file contains various polyfills and other compatibility hacks
 */

// Don't complain about extending native data types -- that's what polyfills do
/* eslint-disable no-extend-native */
