/* SPDX-License-Identifier: LGPL-2.1-or-later */
.pf-v6-c-modal-box__body .scroll {
  max-block-size: calc(75vh - 10rem);
  overflow: auto;
}

.dialog-wait-ct-spinner {
  margin-inline-start: var(--pf-t--global--spacer--sm);
}
