/* SPDX-License-Identifier: LGPL-2.1-or-later */
.contextMenu {
  position: fixed;
  /* xterm accessibility tree has z-index 100 and we need to be in front of it
     * to be able to handle mouse events.
     */
  z-index: 101;
  /* Move the menu under the mouse */
  transform: translate(calc(-1 * var(--pf-t--global--spacer--sm)), calc(-1 * var(--pf-t--global--spacer--sm)));

  &Option .pf-v6-c-menu__item-text {
    display: flex;
    gap: var(--pf-t--global--spacer--sm);
    justify-content: space-between;
    min-inline-size: 10rem;
  }

  &Shortcut {
    opacity: 0.75;
  }
}
